/*--------------------------------------------------------------*/
/* maze.c -- general purpose maze router routines.		*/
/*								*/
/* This file contains the main cost evaluation routine, 	*/
/* the route segment generator, and a routine to search		*/
/* the database for all parts of a routed network.		*/
/*--------------------------------------------------------------*/
/* Written by Tim Edwards, June 2011, based on code by Steve	*/
/* Beccue							*/
/*--------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#define  MAZE

#include "qrouter.h"
#include "config.h"
#include "node.h"
#include "maze.h"

extern int TotalRoutes;

int    CurrentLay;
int    NPX, NPY;

/*--------------------------------------------------------------*/
/* set_node_to_net() ---					*/
/*								*/
/* Change the Obs2[][] matrix values to "newnet" for all tap	*/
/* positions of route terminal "node".  Then follow all routes	*/
/* connected to "node", updating their positions.  Where those	*/
/* routes connect to other nodes, repeat recursively.		*/
/*								*/
/* Return value is 1 if at least one terminal of the node	*/
/* is already marked as "newnet", indicating that the node	*/
/* has already been routed.  Otherwise, the return value is	*/
/* zero of no error occured, and -1 if any point was found to	*/
/* be unoccupied by any net, which should not happen.		*/
/*--------------------------------------------------------------*/

int set_node_to_net(NODE node, int newnet, POINT *pushlist)
{
    int x, y, lay;
    int oldnet;
    int result = 0;
    POINT gpoint;
    DPOINT ntap;
    ROUTE rt;
    SEG seg;
    NODE n2;

    // Process tap points of the node

    for (ntap = node->taps; ntap; ntap = ntap->next) {
       lay = ntap->layer;
       x = ntap->gridx;
       y = ntap->gridy;
       oldnet = Obs2[lay][OGRID(x, y, lay)];
       if (!oldnet) return -1;			// This should not happen.
       if (oldnet == SRCNETNUM) {
	  result = 1;				// Node is already connected!
       }
       else if (oldnet != newnet) {
          Obs2[lay][OGRID(x, y, lay)] = newnet;

	  // push this point on the stack to process

	  if (pushlist != NULL) {
	     gpoint = (POINT)malloc(sizeof(struct point_));
	     gpoint->x1 = x;
	     gpoint->y1 = y;
	     gpoint->layer = lay;
	     gpoint->next = *pushlist;
	     *pushlist = gpoint;
	  }
       }
    }

    // Do the same for point in the halo around the tap, but only if
    // they have been attached to the net during a past routing run.

    for (ntap = node->extend; ntap; ntap = ntap->next) {
       lay = ntap->layer;
       x = ntap->gridx;
       y = ntap->gridy;
       oldnet = Obs2[lay][OGRID(x, y, lay)];
       if (oldnet == SRCNETNUM) {
	  result = 1;				// Node is already connected!
       }
       else if (oldnet == node->netnum) {
          Obs2[lay][OGRID(x, y, lay)] = newnet;

	  // push this point on the stack to process

	  if (pushlist != NULL) {
	     gpoint = (POINT)malloc(sizeof(struct point_));
	     gpoint->x1 = x;
	     gpoint->y1 = y;
	     gpoint->layer = lay;
	     gpoint->next = *pushlist;
	     *pushlist = gpoint;
	  }
       }
    }

    for (rt = node->routes; rt; rt = rt->next) {
       if (rt->segments && rt->node1) {
	  for (seg = rt->segments; seg; seg = seg->next) {
	     lay = seg->layer;
	     x = seg->x1;
	     y = seg->y1;
	     while (1) {
	        oldnet = Obs2[lay][OGRID(x, y, lay)];
		if (!oldnet) break;
		if ((oldnet != newnet) && (oldnet != SRCNETNUM)) {
		   Obs2[lay][OGRID(x, y, lay)] = newnet;

		   // push this point on the stack to process

		   if (pushlist != NULL) {
	  	      gpoint = (POINT)malloc(sizeof(struct point_));
	  	      gpoint->x1 = x;
	  	      gpoint->y1 = y;
	  	      gpoint->layer = lay;
	  	      gpoint->next = *pushlist;
	 	      *pushlist = gpoint;
		   }

		   n2 = Nodeloc[lay][OGRID(x, y, lay)];
		   if ((n2 != (NODE)NULL) && (n2 != node)) {
		      result = set_node_to_net(n2, newnet, pushlist);
		   }
		}

		// Move to next grid position in segment
		if (x == seg->x2 && y == seg->y2) break;
		if (seg->x2 > seg->x1) x++;
		else if (seg->x2 < seg->x1) x--;
		if (seg->y2 > seg->y1) y++;
		else if (seg->y2 < seg->y1) y--;
	     }
	  }
       }
    }
    return result;
}

/*--------------------------------------------------------------*/
/* eval_pt - evaluate cost to get from given pt (x, y, lay) to	*/
/*	current point (NPX, NPY, CurrentLay)			*/
/*	ONLY consider the cost of the single step itself.	*/
/*								*/
/*  ARGS: none							*/
/*  RETURNS: 1 if node needs to be (re)processed, 0 if not.	*/
/*  SIDE EFFECTS: none (get this right or else)			*/
/*--------------------------------------------------------------*/

int eval_pt(int x, int y, int lay, int thisindex)
{
    int k, j, thiscost = 0;
    NODE node;

    k = Obs2[lay][OGRID(x, y, lay)];

    if (!k) {
       /* Not visited before, so create new entry */
       PRind++;

       /* Check bounds on PRind */
       if (PRind >= PRindMAX) return -1;

       /* Create new potential route structure */
       k = PRind;
       Pr[k].pred = 0;
       Pr[k].processed = 0;
       Pr[k].cost = MAXRT;

       if (Verbose > 0) {
	  fprintf(stdout, "PRind = %d at (%d %d %d)\n", k, x, y, lay);
       }

       k |= RTFLAG;
       Obs2[lay][OGRID(x, y, lay)] = k;
    }
    else if (!(k & RTFLAG)) {
       if (k != TARGNETNUM) {
          return 0;		// Position is not routeable
       }
       k = RTFLAG;		// Use Pr[0] to hold target cost info
    }

    // Compute the cost to step from (NPX, NPY, CurrentLay) to (x, y, lay)
    // "BlockCost" is used if the node has only one point to connect to,
    // so that routing over it could block it entirely.

    if (lay > 0) {
	if ((node = Nodeloc[lay - 1][OGRID(x, y, lay - 1)]) != (NODE)NULL) {
	    j = Obs2[lay - 1][OGRID(x, y, lay - 1)];
	    if ((j != TARGNETNUM) && (j != SRCNETNUM)) {
		if (node->taps->next == NULL)
		   thiscost = BlockCost;	// Cost to block out a tap
		else
	           thiscost = XverCost;	// Cross-under cost
	    }
	}
    }
    if (lay < Num_layers - 1) {
	if ((node = Nodeloc[lay + 1][OGRID(x, y, lay + 1)]) != (NODE)NULL) {
	    j = Obs2[lay + 1][OGRID(x, y, lay + 1)];
	    if ((j != TARGNETNUM) && (j != SRCNETNUM)) {
		if (node->taps->next == NULL)
		   thiscost = BlockCost;	// Cost to block out a tap
		else
	           thiscost = XverCost;	// Cross-over cost
	    }
	}
    }
    if (CurrentLay != lay) thiscost += ViaCost;
    if (NPX != x) thiscost += (Vert[lay] * JogCost + (1 - Vert[lay]) * SegCost);
    if (NPY != y) thiscost += (Vert[lay] * SegCost + (1 - Vert[lay]) * JogCost);

    // Add the cost to the cost of the original position
    if (thisindex) {	/* == 0 only for source node */
       thiscost += Pr[thisindex].cost;
    }
   
    // Replace node information if cost is minimum

    k &= ~RTFLAG;
    if (thiscost < Pr[k].cost) {
       Pr[k].pred = thisindex;
       Pr[k].x1 = NPX;
       Pr[k].y1 = NPY;
       Pr[k].layer = CurrentLay;
       Pr[k].cost = thiscost;
       Pr[k].processed = 0;		// Need to reprocess this node

       if (Verbose > 0) {
	  fprintf(stdout, "New cost %d at (%d %d %d)\n", thiscost, x, y, lay);
       }
       return 1;
    }
    return 0;	// New position did not get a lower cost

} /* eval_pt() */

/*--------------------------------------------------------------*/
/* commit_proute - turn the potential route into an actual	*/
/*		route by generating the route segments		*/
/*								*/
/*  ARGS: none							*/
/*  RETURNS: 							*/
/*  SIDE EFFECTS: Obs update, RT llseg added			*/
/*--------------------------------------------------------------*/

int commit_proute(ROUTE rt)
{
   SEG  seg, lseg;
   int  i, j, k;
   int  x, y;
   NODE n1;
   u_int netnum, dir1, dir2;
   u_char first = (u_char)1;
   int  dist, max, xn, yn, ll, dx, dy;
   int  mindist;
   int  layer, lay;
   int  index, prev;
   DPOINT dp;

   fflush(stdout);
   fprintf(stderr, "Commit: TotalRoutes = %d\n", TotalRoutes);

   n1 = rt->node1;
   netnum = rt->netnum;

   xn = NPX; yn = NPY; ll = CurrentLay;
   index = Obs2[ll][OGRID(xn, yn, ll)];

   if (!(index & RTFLAG)) {
   fprintf(stderr, "commit_proute(): impossible - no prouter here!\n");
     print_obs2("obs2.failed");
     exit(-2);
   }

   if (Debug) {
      print_obs("obs.commit");
      print_obs2("obs2.commit");
   }

   index &= ~RTFLAG;
 
   while (1) {
      dist = 1;
      seg = (SEG)malloc(sizeof(struct seg_));
      seg->next = NULL;

      seg->x1 = xn;
      seg->y1 = yn;
      seg->layer = MIN(ll, Pr[index].layer);

      seg->x2 = Pr[index].x1;
      seg->y2 = Pr[index].y1;

      seg->segtype = (ll == Pr[index].layer) ? ST_WIRE : ST_VIA;

      prev = Pr[index].pred;
      dx = seg->x2 - seg->x1;
      dy = seg->y2 - seg->y1;

      // segments are in order---place final segment at end of list
      if (rt->segments == NULL)
	 rt->segments = seg;
      else
	 lseg->next = seg;
      lseg = seg;

      // Continue processing predecessors as long as the direction is the same,
      // so we get a single long wire segment.  This minimizes the number of
      // segments produced.  Vias have to be handled one at a time, as we make
      // no assumptions about stacked vias.

      if (seg->segtype == ST_WIRE) {
	 do {
	    if (((Pr[prev].x1 - Pr[index].x1) == dx) &&
			((Pr[prev].y1 - Pr[index].y1) == dy) &&
			(Pr[prev].layer == Pr[index].layer)) {
	       index = prev;
	       seg->x2 = Pr[index].x1;
	       seg->y2 = Pr[index].y1;
	       prev = Pr[index].pred;
	       dist++;
	    }
	    else
	       break;
	 } while (prev != 0);
      }

      if (Verbose > 0) {
         printf( "commit: index = %d, net = %d, predecessor = %d\n",
		index, netnum, prev);

	 if (seg->segtype == ST_WIRE) {
            printf( "commit: wire layer %d, (%d,%d) to (%d,%d)\n",
		seg->layer, seg->x1, seg->y1, seg->x2, seg->y2);
	 }
	 else {
            printf( "commit: via %d to %d\n", seg->layer, seg->layer + 1);
	 }
	 fflush(stdout);
      }

      // now fill in the Obs structure with this route....

      dir1 = Obs[seg->layer][OGRID(seg->x1, seg->y1, seg->layer)] & PINOBSTRUCTMASK;
      dir2 = Obs[seg->layer][OGRID(seg->x2, seg->y2, seg->layer)] & PINOBSTRUCTMASK;

      if (seg->segtype == ST_VIA) {
	  Obs[seg->layer][OGRID(xn, yn, seg->layer)] = netnum;
	  Obs[seg->layer + 1][OGRID(xn, yn, seg->layer + 1)] = netnum;
      }
      else if (seg->segtype == ST_WIRE) {
	 if (seg->x1 < seg->x2) {
	    for (i = seg->x1; i <= seg->x2; i++) {
	       Obs[ll][OGRID(i, yn, ll)] = netnum;
	    }
	 }
	 if (seg->x1 > seg->x2) {
	    for (i = seg->x2; i <= seg->x1; i++) {
	       Obs[ll][OGRID(i, yn, ll)] = netnum;
	    }
	 }
	 if (seg->y1 < seg->y2) {
	    for (i = seg->y1; i <= seg->y2; i++) {
	       Obs[ll][OGRID(xn, i, ll)] = netnum;
	    }
	 }
	 if (seg->y1 > seg->y2) {
	    for (i = seg->y2; i <= seg->y1; i++) {
	       Obs[ll][OGRID(xn, i, ll)] = netnum;
	    }
	 }
      }
      if (first && dir1) {
	 first = (u_char)0;
	 Obs[seg->layer][OGRID(seg->x1, seg->y1, seg->layer)] |= dir1;
      }

      xn = seg->x2; yn = seg->y2; ll = Pr[index].layer;

      index = prev;

      // Problem:  index should always be zero at source, but it's not.
      // So we check here if the predecessor position is equal to the source.
      // (Is this true any longer?)

      dp = n1->taps;
      if ((index == 0) || ((xn == dp->gridx) && (yn == dp->gridy)
		&& (ll == dp->layer))) { 

         if (dir2) {
	    Obs[seg->layer][OGRID(seg->x2, seg->y2, seg->layer)] |= dir2;
         }

	 if (Debug) print_obs("obs.f1");

	 return TRUE;
      }
   }

   // This block is not reachable

   if (Debug) {
      print_obs( "obs.f2" );
      print_obs2( "obs2.f2" );
   }
   return FALSE;

} /* commit_proute() */

/* end of maze.c */
